@@ -13,6 +13,7 @@ use ckb_db::RocksDB;
 use ckb_db_schema::COLUMNS;
 use ckb_error::{Error, InternalErrorKind};
 use ckb_freezer::Freezer;
+use ckb_logger::{error, info};
 use ckb_notify::{NotifyController, NotifyService, PoolTransactionEntry};
 use ckb_proposal_table::ProposalTable;
 use ckb_proposal_table::ProposalView;
@@ -21,17 +22,19 @@ use ckb_snapshot::{Snapshot, SnapshotMgr};
 use ckb_stop_handler::StopHandler;
 use ckb_store::ChainDB;
 use ckb_store::ChainStore;
-use ckb_tx_pool::{error::Reject, TokioRwLock, TxEntry, TxPool, TxPoolServiceBuilder};
+use ckb_tx_pool::{
+    error::Reject, service::TxVerificationResult, TokioRwLock, TxEntry, TxPool,
+    TxPoolServiceBuilder,
+};
 use ckb_types::core::EpochExt;
 use ckb_types::core::HeaderView;
-use ckb_types::packed::Byte32;
-use ckb_verification::cache::TxVerifyCache;
-use p2p::SessionId as PeerIndex;
+use ckb_verification::cache::init_cache;
 use std::cmp::Ordering;
 use std::collections::HashSet;
-use std::path::PathBuf;
+use std::path::{Path, PathBuf};
 use std::sync::atomic::AtomicBool;
 use std::sync::Arc;
+use tempfile::TempDir;
 
 /// Shared builder for construct new shared.
 pub struct SharedBuilder {
@@ -43,11 +46,13 @@ pub struct SharedBuilder {
     block_assembler_config: Option<BlockAssemblerConfig>,
     notify_config: Option<NotifyConfig>,
     async_handle: Handle,
-    // async stop handle, only test will be assigned
-    async_stop: Option<StopHandler<()>>,
 }
 
-pub fn open_or_create_db(config: &DBConfig) -> Result<RocksDB, ExitCode> {
+pub fn open_or_create_db(
+    bin_name: &str,
+    root_dir: &Path,
+    config: &DBConfig,
+) -> Result<RocksDB, ExitCode> {
     let migrate = Migrate::new(&config.path);
 
     let read_only_db = migrate.open_read_only_db().map_err(|e| {
@@ -73,10 +78,27 @@ pub fn open_or_create_db(config: &DBConfig) -> Result<RocksDB, ExitCode> {
                         You can use the old version CKB if you don't want to do the migration.\n\
                         We strongly recommended you to use the latest stable version of CKB, \
                         since the old versions may have unfixed vulnerabilities.\n\
-                        Run `ckb migrate --help` for more information about migration."
+                        Run `\"{}\" migrate -C \"{}\"` and confirm by typing \"YES\" to migrate the data.\n\
+                        We strongly recommend that you backup the data directory before migration.",
+                        bin_name,
+                        root_dir.display()
                     );
                     Err(ExitCode::Failure)
                 } else {
+                    info!("process fast migrations ...");
+
+                    let bulk_load_db_db = migrate.open_bulk_load_db().map_err(|e| {
+                        eprintln!("migrate error {}", e);
+                        ExitCode::Failure
+                    })?;
+
+                    if let Some(db) = bulk_load_db_db {
+                        migrate.migrate(db).map_err(|err| {
+                            eprintln!("Run error: {:?}", err);
+                            ExitCode::Failure
+                        })?;
+                    }
+
                     Ok(RocksDB::open(config, COLUMNS))
                 }
             }
@@ -94,11 +116,13 @@ pub fn open_or_create_db(config: &DBConfig) -> Result<RocksDB, ExitCode> {
 impl SharedBuilder {
     /// Generates the base SharedBuilder with ancient path and async_handle
     pub fn new(
+        bin_name: &str,
+        root_dir: &Path,
         db_config: &DBConfig,
         ancient: Option<PathBuf>,
         async_handle: Handle,
     ) -> Result<SharedBuilder, ExitCode> {
-        let db = open_or_create_db(db_config)?;
+        let db = open_or_create_db(bin_name, root_dir, db_config)?;
 
         Ok(SharedBuilder {
             db,
@@ -109,24 +133,52 @@ impl SharedBuilder {
             store_config: None,
             block_assembler_config: None,
             async_handle,
-            async_stop: None,
         })
     }
 
     /// Generates the SharedBuilder with temp db
     pub fn with_temp_db() -> Self {
-        let (handle, stop) = new_global_runtime();
-        SharedBuilder {
-            db: RocksDB::open_tmp(COLUMNS),
+        use once_cell::{sync, unsync};
+        use std::{
+            borrow::Borrow,
+            sync::atomic::{AtomicUsize, Ordering},
+        };
+
+        // once #[thread_local] is stable
+        // #[thread_local]
+        // static RUNTIME_HANDLE: unsync::OnceCell<...
+
+        thread_local! {
+            static TMP_DIR: sync::OnceCell<TempDir> = sync::OnceCell::new();
+            static RUNTIME_HANDLE: unsync::OnceCell<(Handle, StopHandler<()>)> = unsync::OnceCell::new();
+        }
+
+        static DB_COUNT: AtomicUsize = AtomicUsize::new(0);
+
+        let db = {
+            let db_id = DB_COUNT.fetch_add(1, Ordering::SeqCst);
+            let db_base_dir = TMP_DIR.with(|tmp_dir| {
+                tmp_dir
+                    .borrow()
+                    .get_or_try_init(TempDir::new)
+                    .unwrap()
+                    .path()
+                    .to_path_buf()
+            });
+            let db_dir = db_base_dir.join(format!("db_{}", db_id));
+            RocksDB::open_in(db_dir, COLUMNS)
+        };
+
+        RUNTIME_HANDLE.with(|runtime| SharedBuilder {
+            db,
             ancient_path: None,
             consensus: None,
             tx_pool_config: None,
             notify_config: None,
             store_config: None,
             block_assembler_config: None,
-            async_handle: handle,
-            async_stop: Some(stop),
-        }
+            async_handle: runtime.borrow().get_or_init(new_global_runtime).0.clone(),
+        })
     }
 }
 
@@ -261,7 +313,6 @@ impl SharedBuilder {
             block_assembler_config,
             notify_config,
             async_handle,
-            async_stop,
         } = self;
 
         let tx_pool_config = tx_pool_config.unwrap_or_else(Default::default);
@@ -276,9 +327,7 @@ impl SharedBuilder {
             ExitCode::Failure
         })?;
 
-        let txs_verify_cache = Arc::new(TokioRwLock::new(TxVerifyCache::new(
-            tx_pool_config.max_verify_cache_size,
-        )));
+        let txs_verify_cache = Arc::new(TokioRwLock::new(init_cache()));
 
         let (snapshot, table) =
             Self::init_snapshot(&store, Arc::clone(&consensus)).map_err(|e| {
@@ -295,7 +344,6 @@ impl SharedBuilder {
             Arc::clone(&snapshot),
             block_assembler_config,
             Arc::clone(&txs_verify_cache),
-            Arc::clone(&snapshot_mgr),
             &async_handle,
             sender.clone(),
         );
@@ -311,7 +359,6 @@ impl SharedBuilder {
             consensus,
             snapshot_mgr,
             async_handle,
-            async_stop,
             ibd_finished,
             sender,
         );
@@ -331,7 +378,7 @@ impl SharedBuilder {
 pub struct SharedPackage {
     table: Option<ProposalTable>,
     tx_pool_builder: Option<TxPoolServiceBuilder>,
-    relay_tx_receiver: Option<Receiver<(Option<PeerIndex>, Byte32)>>,
+    relay_tx_receiver: Option<Receiver<TxVerificationResult>>,
 }
 
 impl SharedPackage {
@@ -346,7 +393,7 @@ impl SharedPackage {
     }
 
     /// Takes the relay_tx_receiver out of the package, leaving a None in its place.
-    pub fn take_relay_tx_receiver(&mut self) -> Receiver<(Option<PeerIndex>, Byte32)> {
+    pub fn take_relay_tx_receiver(&mut self) -> Receiver<TxVerificationResult> {
         self.relay_tx_receiver
             .take()
             .expect("take relay_tx_receiver")
@@ -383,6 +430,7 @@ fn register_tx_pool_callback(tx_pool_builder: &mut TxPoolServiceBuilder, notify:
             cycles: entry.cycles,
             size: entry.size,
             fee: entry.fee,
+            timestamp: entry.timestamp,
         };
         notify_pending.notify_new_transaction(notify_tx_entry);
     }));
@@ -401,6 +449,7 @@ fn register_tx_pool_callback(tx_pool_builder: &mut TxPoolServiceBuilder, notify:
                 cycles: entry.cycles,
                 size: entry.size,
                 fee: entry.fee,
+                timestamp: entry.timestamp,
             };
             notify_proposed.notify_proposed_transaction(notify_tx_entry);
         },
@@ -416,12 +465,23 @@ fn register_tx_pool_callback(tx_pool_builder: &mut TxPoolServiceBuilder, notify:
             // update statics
             tx_pool.update_statics_for_remove_tx(entry.size, entry.cycles);
 
+            let tx_hash = entry.transaction().hash();
+            // record recent reject
+            if matches!(reject, Reject::Resolve(..)) {
+                if let Some(ref mut recent_reject) = tx_pool.recent_reject {
+                    if let Err(e) = recent_reject.put(&tx_hash, reject.clone()) {
+                        error!("record recent_reject failed {} {} {}", tx_hash, reject, e);
+                    }
+                }
+            }
+
             // notify
             let notify_tx_entry = PoolTransactionEntry {
                 transaction: entry.rtx.transaction.clone(),
                 cycles: entry.cycles,
                 size: entry.size,
                 fee: entry.fee,
+                timestamp: entry.timestamp,
             };
             notify_reject.notify_reject_transaction(notify_tx_entry, reject);
         },
